"
    Copyright (C) 2016-now Jan Vrany

    This code is not an open-source (yet). You may use this code
    for your own experiments and projects, given that:

    * all modification to the code will be sent to the
      original author for inclusion in future releases
    * this is not used in any commercial software

    This license is provisional and may (will) change in
    a future.
"
"{ Package: 'jv:dragonfly/c1' }"

"{ NameSpace: DragonFly }"

Object subclass:#C1Compiler
	instanceVariableNames:'method module function asm prologue epilogue'
	classVariableNames:'SelectorSpecialCharMappingTable'
	poolDictionaries:'DragonFly::C1LLVMMTypes LLVMIntPredicate VMData VMOffsets'
	category:'DragonFly-C1'
!

!C1Compiler class methodsFor:'documentation'!

copyright
"
    Copyright (C) 2016-now Jan Vrany

    This code is not an open-source (yet). You may use this code
    for your own experiments and projects, given that:

    * all modification to the code will be sent to the
      original author for inclusion in future releases
    * this is not used in any commercial software

    This license is provisional and may (will) change in
    a future.

"
! !

!C1Compiler class methodsFor:'class initialization'!

initialize
    "Invoked at system start or when the class is dynamically loaded."

    "/ please change as required (and remove this comment)

    SelectorSpecialCharMappingTable := Dictionary withKeysAndValues:
                #($+ 'pl'
                  $- 'mi'
                  $* 'mu'
                  $/ 'di'
                  $, 'co'
                  $@ 'at'
                  $< 'le'
                  $> 'gr'
                  $= 'eq'
                  $~ 'ne'
                  $| 'pi'
                  $\ 'mo'
                  $& 'am').

    "Created: / 09-02-2016 / 08:43:21 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1Compiler class methodsFor:'compilation'!

compile: method
    ^ self new compile: method

    "Created: / 09-02-2016 / 08:47:10 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1Compiler class methodsFor:'utilities'!

functionNameForClass: class selector: selector      
    "For given class name and selector, returns the name 
     used by LLVM"

    ^ String streamContents:[ :s|
        s nextPutAll: '__M_D_'.
        s nextPutAll: (class name copyReplaceAll: $: with: $_).
        s nextPut: $_.
        selector isBinarySelector ifTrue:[ 
            selector do:[:c |     
                s nextPutAll: (SelectorSpecialCharMappingTable at: c)
            ].
        ] ifFalse:[ 
            selector do:[:c |  
                c isAlphaNumeric ifTrue:[ 
                    s nextPut: c 
                ] ifFalse:[
                    s nextPut: $_.
                    c == $: ifFalse:[ 
                        c codePoint printOn: s.
                    ]
                ].
            ]
        ].
    ].

    "
    C1Compiler functionNameForClass: C1Compiler selector: #functionNameForClass:selector: 
    C1Compiler functionNameForClass: SmallInteger selector: #+
    C1Compiler functionNameForClass: Object selector: #~=
    C1Compiler functionNameForClass: JavaVM selector: #_ALOAD:

    "

    "Created: / 09-02-2016 / 08:44:10 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1Compiler methodsFor:'accessing'!

method
    ^ method
!

method:aMethod
    method := aMethod.
! !

!C1Compiler methodsFor:'compilation'!

compile: aMethod
    self method: aMethod.
    self prepare.
    self prologue.

    self epilogue.
    self finish.

    "Created: / 09-02-2016 / 08:52:42 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 21-04-2016 / 09:18:35 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

finish
    "Finish the compilation, i.e., generate machine code
     and update method's code pointer"
    | opts mm object jit |

    object := CompiledCodeObject forCompiledCode: method text: 0 literals: 0 ilcs: 0.
    mm := C1LLVMMCJITMemoryManager for: object.
    opts := LLVMMCJITCompilerOptions new.
    opts MCJMM: mm.
    jit := LLVMExecutionEngine newForModule: module options: opts.
    method code: (jit addressOfFunction: function).

    "Created: / 21-04-2016 / 09:19:12 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

prepare
    | name |
    name := self class functionNameForClass: method mclass ? UndefinedObject selector: method selector ? '<unbound>'.
    module := LLVMModule newWithName: name.
    function := module addFunctionNamed: name type: (TyOBJFUNCs at: method numArgs + 1).
    (function parameterAt: OBJFUNCArgIndexReceiver) name: 'zelf'.
    (function parameterAt: OBJFUNCArgIndexSelector) name: 'selector'.
    (function parameterAt: OBJFUNCArgIndexSearchClass) name: 'searchClass'.
    (function parameterAt: OBJFUNCArgIndexILC) name: 'pilc'.
    1 to: method numArgs do:[:i | (function parameterAt: OBJFUNCArgIndexArgBase + i) name: 'arg' , i printString ].
    asm := function builder.

    "Created: / 21-04-2016 / 09:17:06 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 21-04-2016 / 16:52:16 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1Compiler methodsFor:'private'!

epilogue
    self halt.

    "Created: / 15-04-2016 / 23:38:09 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

prologue
    | classOfReceiver classInILC classCheckFailedBlock classCheckPassedBlock classCheck |
    prologue := function addBasicBlockNamed: 'prologue'.
    asm continue: prologue.
    classOfReceiver := self loadClassOf:(function parameterAt:OBJFUNCArgIndexReceiver).
    classInILC := self loadClassFromILC:(function parameterAt:OBJFUNCArgIndexILC).
    classCheckFailedBlock := function addBasicBlockNamed: 'prologue.class-check-failed'. 
    classCheckPassedBlock := function addBasicBlockNamed: 'prologue.class-check-passed'.
    classCheck := asm icmp: (asm ptr: classOfReceiver toInt: LLVMType intptr)
                         _: (asm ptr: classInILC toInt: LLVMType intptr)
                      cond: LLVMIntEQ.
    asm if: classCheck then: classCheckPassedBlock else: classCheckFailedBlock.

    "/ Class check failed, call _SENDX()"
    asm block: classCheckFailedBlock.


    "/ Classes match, continue.
    asm block: classCheckPassedBlock.

    "Created: / 09-02-2016 / 17:07:28 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 21-04-2016 / 12:05:01 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1Compiler methodsFor:'private-fetch / store'!

loadClassFromILC:pIlc 
    self assert:pIlc type = TyInlineCachePtr.
    ^ asm load:(asm gep:pIlc at:{ 0 . TyInlineCacheIndexClass })
             "Created: / 12-02-2016 / 13:19:19 / Jan Vrany <jan.vrany@fit.cvut.cz>" "Modified: / 15-04-2016 / 23:18:52 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadClassOf:obj
    | classes |

    classes := Array new: 2.
    asm if: (self isSmallIntegerObject:obj) then: [
        classes at: 1 put: self loadClassSmallInteger.
    ] else:[  
        classes at: 2 put: (asm load:(asm gep:obj at:{ 0 . TyInstanceFieldIndexClass })).
    ].
    ^ asm phi: classes

    "Created: / 09-02-2016 / 17:25:48 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 22-04-2016 / 15:11:51 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadClassSmallInteger
    | addr |

    addr := asm 
                int:(LLVMConstant uintptr:ADDR_SmallInteger)
                toPtr:TyOBJ pointer
                as:'ADDR_SmallInteger'.
    ^ asm load:addr as: 'SmallInteger'.

    "Created: / 12-02-2016 / 11:54:10 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 21-04-2016 / 17:03:02 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadNil
    ^ asm int:(LLVMConstant uintptr:0) toPtr:TyOBJ.

    "Created: / 21-04-2016 / 13:59:08 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1Compiler methodsFor:'private-testing'!

isSmallIntegerObject: value
    self assert: value type = TyOBJ.
    ^ asm icmp: (asm and: (asm ptr: value toInt: LLVMType intptr) _: (LLVMConstant uintptr: 1))
             _: (LLVMConstant uintptr: 1)
          cond: LLVMIntEQ

    "Created: / 09-02-2016 / 17:30:55 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 12-02-2016 / 11:55:11 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1Compiler class methodsFor:'documentation'!

version_HG

    ^ '$Changeset: <not expanded> $'
! !


C1Compiler initialize!
